'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var rxjs = require('rxjs');
var React = require('react');

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */
/* global Reflect, Promise */

var extendStatics = Object.setPrototypeOf ||
    ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
    function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };

function __extends(d, b) {
    extendStatics(d, b);
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
}

var __assign = Object.assign || function __assign(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
    }
    return t;
};

var PROPS_EFFECT = '@@refract/effect/props';
var toProps = function (props) { return ({
    type: PROPS_EFFECT,
    payload: {
        replace: false,
        props: props
    }
}); };
var asProps = function (props) { return ({
    type: PROPS_EFFECT,
    payload: {
        replace: true,
        props: props
    }
}); };

var subscribeToSink = function (sink, next, error) {
    return sink.subscribe({
        next: next,
        error: error
    });
};
var createObservable = function (subscribe) {
    return rxjs.Observable.create(function (Listener) {
        var unsubscribe = subscribe(Listener);
        return unsubscribe;
    });
};

var shallowEquals = function (left, right) {
    return left === right ||
        (Object.keys(left).length === Object.keys(right).length &&
            Object.keys(left).every(function (leftKey) { return left[leftKey] === right[leftKey]; }));
};
var configureComponent = function (handler, errorHandler) { return function (aperture, instance, isValidElement, isComponentClass) {
    if (isValidElement === void 0) { isValidElement = function () { return false; }; }
    if (isComponentClass === void 0) { isComponentClass = function () { return false; }; }
    instance.state = {
        children: null,
        props: {},
        decoratedProps: {}
    };
    var setState = function (state) {
        if (instance.unmounted) {
            return;
        }
        if (instance.mounted) {
            instance.setState(state);
        }
        else {
            instance.state = state;
        }
    };
    var finalHandler = function (initialProps) {
        var effectHandler = handler(initialProps);
        return function (effect) {
            if (isValidElement(effect)) {
                setState({
                    children: effect
                });
            }
            else if (effect && effect.type === PROPS_EFFECT) {
                var payload_1 = effect.payload;
                setState({
                    replace: payload_1.replace,
                    props: payload_1.props,
                    decoratedProps: Object.keys(payload_1.props || {}).reduce(function (props, propName) {
                        var prop = payload_1.props[propName];
                        var previousProp = instance.state.props[propName];
                        if (typeof prop === 'function' &&
                            prop !== previousProp) {
                            decorateProp(props, prop, propName);
                        }
                        return props;
                    }, {})
                });
            }
            else {
                effectHandler(effect);
            }
        };
    };
    var listeners = {
        mount: [],
        unmount: [],
        allProps: [],
        props: {},
        fnProps: {},
        event: {}
    };
    var decoratedProps = {};
    var pushEvent = function (eventName) { return function (val) {
        (listeners.event[eventName] || []).forEach(function (listener) {
            return listener.next(val);
        });
    }; };
    var decorateProp = function (container, prop, propName) {
        if (propName === 'children' || isComponentClass(prop)) {
            return;
        }
        container[propName] = function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            (listeners.fnProps[propName] || []).forEach(function (l) { return l.next(args[0]); });
            return prop.apply(void 0, args);
        };
    };
    Object.keys(instance.props).forEach(function (propName) {
        if (typeof instance.props[propName] === 'function') {
            decorateProp(decoratedProps, instance.props[propName], propName);
        }
    });
    var mountObservable = createObservable(function (listener) {
        listeners.mount = listeners.mount.concat(listener);
        return function () { return listeners.mount.filter(function (l) { return l !== listener; }); };
    });
    var unmountObservable = createObservable(function (listener) {
        listeners.unmount = listeners.unmount.concat(listener);
        return function () { return listeners.unmount.filter(function (l) { return l !== listener; }); };
    });
    var createPropObservable = function (propName) {
        var listenerType = propName
            ? typeof instance.props[propName] === 'function'
                ? 'fnProps'
                : 'props'
            : 'allProps';
        return createObservable(function (listener) {
            if (listenerType === 'allProps') {
                listener.next(instance.props);
                listeners.allProps = listeners.allProps.concat(listener);
                return function () {
                    listeners.allProps.filter(function (l) { return l !== listener; });
                };
            }
            if (listenerType === 'props') {
                listener.next(instance.props[propName]);
            }
            listeners[listenerType][propName] = (listeners[listenerType][propName] || []).concat(listener);
            return function () {
                listeners[listenerType][propName].filter(function (l) { return l !== listener; });
            };
        });
    };
    var createEventObservable = function (eventName) {
        return createObservable(function (listener) {
            listeners.event[eventName] = (listeners.event[eventName] || []).concat(listener);
            return function () {
                listeners.event[eventName].filter(function (l) { return l !== listener; });
            };
        });
    };
    var component = {
        mount: mountObservable,
        unmount: unmountObservable,
        observe: createPropObservable,
        event: createEventObservable,
        pushEvent: pushEvent
    };
    var sinkObservable = aperture(instance.props)(component);
    var sinkSubscription = subscribeToSink(sinkObservable, finalHandler(instance.props), errorHandler ? errorHandler(instance.props) : undefined);
    var sendNext = function (prevProps) {
        Object.keys(listeners.props).forEach(function (propName) {
            var prop = instance.props[propName];
            if (!prevProps || prevProps[propName] !== prop) {
                listeners.props[propName].forEach(function (l) { return l.next(prop); });
            }
        });
        listeners.allProps.forEach(function (l) { return l.next(instance.props); });
    };
    instance.reDecorateProps = function (nextProps) {
        Object.keys(nextProps).forEach(function (propName) {
            if (typeof instance.props[propName] === 'function' &&
                nextProps[propName] !== instance.props[propName]) {
                decorateProp(decoratedProps, nextProps[propName], propName);
            }
        });
    };
    instance.pushProps = function (prevProps) {
        sendNext(prevProps);
    };
    instance.triggerMount = function () {
        listeners.mount.forEach(function (l) { return l.next(undefined); });
    };
    instance.triggerUnmount = function () {
        listeners.unmount.forEach(function (l) { return l.next(undefined); });
        sinkSubscription.unsubscribe();
    };
    instance.getChildProps = function () {
        var state = instance.state;
        var stateProps = __assign({}, state.props, state.decoratedProps);
        if (state.replace === true) {
            return __assign({}, stateProps, { pushEvent: pushEvent });
        }
        var componentProps = __assign({}, instance.props, decoratedProps, { pushEvent: pushEvent });
        if (state.replace === false) {
            return __assign({}, componentProps, stateProps);
        }
        return componentProps;
    };
    instance.havePropsChanged = function (newProps, newState) {
        var state = instance.state;
        if (state.children) {
            return state.children !== newState.children;
        }
        var haveStatePropsChanged = !shallowEquals(state.props, newState.props);
        if (newState.replace === true) {
            return haveStatePropsChanged;
        }
        var havePropsChanged = !shallowEquals(instance.props, newProps);
        if (newState.replace === false) {
            return havePropsChanged || haveStatePropsChanged;
        }
        return havePropsChanged;
    };
}; };

var isComponentClass = function (ComponentClass) {
    return Boolean(ComponentClass &&
        ComponentClass.prototype &&
        ComponentClass.prototype.isReactComponent);
};
var Empty = function () { return null; };
var withEffects = function (handler, errorHandler) { return function (aperture) { return function (BaseComponent) {
    if (BaseComponent === void 0) { BaseComponent = Empty; }
    return /** @class */ (function (_super) {
        __extends(WithEffects, _super);
        function WithEffects(props, context) {
            var _this = _super.call(this, props, context) || this;
            _this.mounted = false;
            _this.unmounted = false;
            configureComponent(handler, errorHandler)(aperture, _this, React.isValidElement, isComponentClass);
            return _this;
        }
        WithEffects.prototype.componentDidMount = function () {
            this.mounted = true;
            this.triggerMount();
        };
        WithEffects.prototype.componentWillReceiveProps = function (nextProps) {
            this.reDecorateProps(nextProps);
        };
        WithEffects.prototype.shouldComponentUpdate = function (nextProps, nextState) {
            return this.havePropsChanged(nextProps, nextState);
        };
        WithEffects.prototype.componentDidUpdate = function (prevProps) {
            this.pushProps(prevProps);
        };
        WithEffects.prototype.componentWillUnmount = function () {
            this.unmounted = true;
            this.triggerUnmount();
        };
        WithEffects.prototype.render = function () {
            if (this.state.children) {
                return this.state.children;
            }
            return React.createElement(BaseComponent, this.getChildProps());
        };
        return WithEffects;
    }(React.Component));
}; }; };

var identity = function (arg) { return arg; };
var compose = function () {
    var fns = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        fns[_i] = arguments[_i];
    }
    if (fns.length === 0) {
        return identity;
    }
    if (fns.length === 1) {
        return fns[0];
    }
    return function (arg) {
        return fns.reduceRight(function (previousResult, fn) { return fn(previousResult); }, arg);
    };
};

exports.withEffects = withEffects;
exports.compose = compose;
exports.asProps = asProps;
exports.toProps = toProps;
exports.PROPS_EFFECT = PROPS_EFFECT;
